/************************************************************************/
/*  mstcorr:								*/
/*	Fix (ie remove) time tears larger than the specified gap.	*/
/*	Optionally set the leapsecond flag.				*/
/*									*/
/*	Douglas Neuhauser						*/
/*	Seismological Laboratory					*/
/*	University of California, Berkeley				*/
/*	doug@seismo.berkeley.edu					*/
/*									*/
/************************************************************************/

/*
 * Copyright (c) 1999-2002 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#ifndef lint
static char sccsid[] = "$Id: mstcorr.c,v 1.3 2002/11/22 23:29:19 doug Exp $ ";
#endif

#include <stdio.h>
#include "version.h"

#define	QUOTE(x)	#x
#define	STRING(x)	QUOTE(x)

#ifndef	DEFAULT_MINGAP
#define	DEFAULT_MINGAP	1000
#endif
#ifndef	DEFAULT_MAXGAP
#define	DEFAULT_MAXGAP	2000000
#endif

char *syntax[] = {
"%s version " VERSION ,
"%s -  Fix time tears larger than the specified time gap.",
"Syntax:",
"%s    [-U] [-g nd][ [-d n] [-h] [input_file [output_file]] | files...",
"    where:",
"	-U	    Update input file instead of creating new output file.",
"		    If you use this option, the file list is a list of",
"		    input/output files instead of an optional input and",
"		    output file.",
"	-g usecs    Minimum gap to correct, specified in usecs (microseconds).",
"		    Default value is " STRING(DEFAULT_MINGAP) " usecs.",
"	-G usecs    Maximum gap to correct, specified in usecs (microseconds).",
"		    A value of 0 specifies infinite maximum value.",
"		    Default value is " STRING(DEFAULT_MAXGAP) " usecs.",
"	-l	    Set leapsecond flag if there is a leapsecond.",
"	-d debug    Value of debug flag (may be ORed together)",
"		    1 = DEBUG_REF",
"	-h	    Help - prints syntax message.",
"	input_file",
"		    Name of the input file to be corrected.",
"		    If no file is specified, read from stdin.",
"	output_file",
"		    Name of the output file to be created.",
"		    If no file is specified, write to stdout.",
"	files...    A list of files to be synced and rewritten.",
"		    Used with the -U option.",
"Notes:",
"1.  The time correction will be added to the MiniSEED time correction field",
"    in the MiniSEED header.  If the time correction flag is set in the",
"    activity flag, the time correction will also be added to the header time.",
"Examples:",
"1.  Fix time tears in a file, writing to a new file.",
"    %s BKS.BK.BHN.D.1997.200 BKS.BK.BHN.D.1997.200.new",
"2.  Fix time tears greater than .5 msec, replacing the original file.",
"    %s -U -g 500 BKS.BK.HH?.D.1997.200",
NULL };

#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#ifdef	SUNOS4
char *dirname  (char *path);
#else
#include <libgen.h>
#endif

#include <qlib2.h>

#define ERROR		-2
#define MAXBLKSIZE	32768
#define	FIXED_DATA_HDR_SIZE 48
#define	DEBUG(x)	(debug_flag&(x))
#define	DEBUG_FIX	1

char *cmdname;
char *in_file = NULL;
char *out_file = NULL;
FILE *info;
int date_fmt = JULIANC_FMT;
int update_flag = 0;
int debug_flag = 0;
int mingap = DEFAULT_MINGAP;
int maxgap = DEFAULT_MAXGAP;
int set_leapsecond_flag;

/************************************************************************/
/* main program.							*/
/************************************************************************/
main (int argc, char **argv) 
{
    FILE *fp_ref;
    FILE *fp_in;
    FILE *fp_out;
    int status;

    /* Variables needed for getopt. */
    extern char	*optarg;
    extern int	optind, opterr;
    int		c;

    /*	Parse command line options.					*/
    cmdname = tail(argv[0]);
    info = stdout;
    fp_in = stdin;
    fp_out = stdout;
    while ( (c = getopt(argc,argv,"hUd:g:G:l")) != -1)
	switch (c) {
	case '?':
	case 'h':   print_syntax(cmdname,syntax,info); exit(0); break;
	case 'U':   update_flag = 1; break;
	case 'd':   debug_flag = atoi(optarg); break;
	case 'g':   mingap = atoi(optarg); break;
	case 'G':   maxgap = atoi(optarg); break;
	case 'l':   set_leapsecond_flag = 1; break;
	default:    fprintf (stderr, "Invalid option: -%c\n",c);
		    exit(1);
	}

    /*	Skip over all options and their arguments.			*/
    argv = &(argv[optind]);
    argc -= optind;

    /* Parse filelist if update_flag is not set.			*/
    if (!update_flag) {
	in_file = "<stdin>";
	if (--argc >= 0) {
	    in_file = *argv++;
	    if ((fp_in = fopen (in_file, "r")) == NULL) {
		fprintf (stderr, "Unable to open input file %s\n", in_file);
		exit(1);
	    }
	}

	out_file = "<stdout>";
	if (--argc >= 0) {
	    out_file = *argv++;
	    if ((fp_out = fopen (out_file, "w")) == NULL) {
		fprintf (stderr, "Unable to open output file %s\n", out_file);
		exit(1);
	    }
	}
	else info = stderr;
	status = fix_file (fp_in, fp_out);
    }

    else {
	/* Process each input file */
	char *pathcopy, *dir;
	while (argc-- > 0) {
	    in_file = *argv++;
	    if ((fp_in = fopen (in_file, "r")) == NULL) {
		fprintf (stderr, "Unable to open input file %s\n", in_file);
		exit(1);
	    }
	    pathcopy = strdup(in_file);
	    dir = dirname (pathcopy);
	    out_file = tempnam (dir, cmdname);
	    free (pathcopy);
	    if ((fp_out = fopen (out_file, "w")) == NULL) {
		fprintf (stderr, "Unable to open output file %s\n", out_file);
		exit(1);
	    }
	    status = fix_file (fp_in, fp_out);
	    if (status != 0) {
		fprintf (stderr, "Error syncing file %s\n", in_file);
		exit (status);
	    }
	    fclose (fp_in);
	    fclose (fp_out);
	    status = rename (out_file, in_file);
	    if (status != 0) {
		fprintf (stderr, "Error renaming output file %s to %s - %s\n",
			 out_file, in_file, strerror (errno));
		exit (1);
	    }
	}
    }
    return (0);
}

/************************************************************************/
/*  fix_file:								*/
/*	Fix time gaps in input file, output fixed file.			*/
/************************************************************************/
int fix_file (FILE *fp_in, FILE *fp_out)
{
    char buf[MAXBLKSIZE];
    char out_buf[MAXBLKSIZE];
    char *p_buf;
    DATA_HDR *hdr, *prev;
    INT_TIME expect, it;
    EXT_TIME et;
    int corr_usecs, corr_ticks;
    int data_offset, datalen;
    int status;
    int seconds, usecs;
    double dusecs, abs_dusecs;

    prev = hdr = NULL;
    /********************************************************************/
    /* loop over records in the input file.				*/
    /********************************************************************/
    while (1) {
        p_buf = buf;
	status = read_ms_record (&hdr, &p_buf, fp_in);
	if (status < 0) break;

	if (prev != NULL) {
	    time_interval2 (prev->num_samples, prev->sample_rate, 
			    prev->sample_rate_mult,
			    &seconds, &usecs);
	    expect = add_time (prev->begtime, seconds, usecs);
	    dusecs = tdiff (expect, hdr->begtime);
	    abs_dusecs = fabs(dusecs);
	    if (abs_dusecs > 0 && abs_dusecs >= mingap && 
		(maxgap <= 0 || abs_dusecs <= maxgap)) {
		/* Compute desired time correction in ticks.*/
		corr_ticks = roundoff (dusecs/USECS_PER_TICK);
		corr_usecs = corr_ticks * USECS_PER_TICK;
		if (DEBUG(DEBUG_FIX)) {
		    fprintf (info, "begtime=%s ", time_to_str(hdr->begtime,date_fmt));
		    fprintf (info, "exptime=%s corr=%d\n", time_to_str(expect,date_fmt),
			     corr_ticks);
		    fflush (info);
		}
		/* Add correction to begtime. */
		hdr->begtime = add_time (hdr->begtime, 0, corr_usecs);
		hdr->endtime = add_time (hdr->endtime, 0, corr_usecs);
		/* Add correction to time correction field. */
		hdr->num_ticks_correction += corr_ticks;
		if (hdr->activity_flags & ACTIVITY_TIME_CORR_APPLIED) {
		    hdr->hdrtime = hdr->begtime;
		}
	    }
	}

	if (set_leapsecond_flag) {
	    it = hdr->begtime;
	    while (tdiff (hdr->endtime,it) > 0) {
		et = int_to_ext(it);
		if (et.second >= 60) {
		    if (DEBUG(DEBUG_FIX)) {
			fprintf (info, "begtime=%s ", time_to_str(hdr->begtime,date_fmt));
			fprintf (info, "endtime=%s leapsecond found\n", time_to_str(hdr->endtime,date_fmt));
			fflush (info);
		    }
		    hdr->activity_flags |= ACTIVITY_POS_LEAP_SECOND;
		}
		it = add_time (it, 1, 0);
	    }
	}

	/* Write the new record to the output. */
	data_offset = hdr->first_data;
	datalen = hdr->blksize - data_offset;
	memset ((void *)out_buf, 0, hdr->blksize);
	init_sdr_hdr ((SDR_HDR *)out_buf, hdr, NULL);
	if (datalen > 0) {
	    memcpy (out_buf+data_offset, buf+data_offset, datalen);
	}
	update_sdr_hdr ((SDR_HDR *)out_buf, hdr);
	status = fwrite (out_buf, hdr->blksize, 1, fp_out);
	if (status == 0) {
	    fprintf (stderr, "Error writing file %s\n", out_file);
	    exit (1);
	}
	if (prev) free_data_hdr (prev);
	prev = hdr;
    }
    return (0);
}
